# -*- coding: utf-8 -*-

"""
Helpers
=======

"""

# %% IMPORTS
# Package imports
import numpy as np

# All declaration
__all__ = ['is_buffer_obj']


# %% FUNCTION DEFINITIONS
# This function checks whether a provided object exposes its internal buffer
def is_buffer_obj(obj):
    """
    Checks if the provided `obj` exposes its internal buffer and can be used in
    uppercase communication methods.
    Currently, only NumPy arrays are seen as buffer objects.

    """

    # Check if provided obj is a NumPy array
    return(isinstance(obj, np.ndarray))



# -*- coding: utf-8 -*-

"""
Utilities
=========
Provides several useful utility functions.

"""


# %% IMPORTS
# MPI import
from mpi4pyd import MPI

# All declaration
__all__ = ['rprint']

# Determine MPI size and ranks
size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()


# %% FUNCTION DEFINITIONS
# Redefine the print function to include the MPI rank if MPI is used
def rprint(*args, **kwargs):
    """
    Custom :func:`~print` function that prepends the rank of the MPI process
    that calls it to the message if the size of the world intra-communicator is
    more than 1.
    Takes the same input arguments as the normal :func:`~print` function.

    """

    # If MPI is used and size > 1, prepend rank to message
    if(MPI.__package__ == 'mpi4py' and size > 1):
        args = list(args)
        args.insert(0, "Rank %i:" % (rank))
    print(*args, **kwargs)



# -*- coding: utf-8 -*-

# %% IMPORTS
# Package imports
import numpy as np
import pytest

# mpi4pyd imports
from mpi4pyd.dummyMPI import Comm, Intracomm, COMM_WORLD as comm


# %% CUSTOM CLASSES
class CommTest(Comm):
    pass


# %% PYTEST CLASSES AND FUNCTIONS
# Pytest for custom Comm class
def test_CommTest():
    test_comm = CommTest()
    assert test_comm.name == 'dummyMPI_CommTest'


# Pytest for custom Intracomm class
def test_Intracomm():
    with pytest.raises(TypeError):
        Intracomm(1)


# Pytest for COMM_WORLD instance
class Test_COMM_WORLD(object):
    array = np.array([1, 2, 3, 4, 5])
    buffer = np.empty_like(array)

    def test_props(self):
        assert comm.Get_name() == 'dummyMPI_COMM_WORLD'
        assert comm.Get_size() == 1
        assert comm.Get_rank() == 0

    def test_Allgather(self):
        comm.Allgather(self.array, self.buffer)
        assert (self.buffer == self.array).all()
        assert (comm.allgather(self.array)[0] == self.array).all()
        comm.Allgatherv(self.array, self.buffer)
        assert (self.buffer == self.array).all()

    def test_Allreduce(self):
        comm.Allreduce(self.array, self.buffer)
        assert (self.buffer == self.array).all()
        assert (comm.allreduce(self.array) == self.array).all()
        assert (comm.allreduce(1) == 1)
        assert comm.reduce([tuple(self.array.tolist())]) ==\
            tuple(self.array.tolist())

    def test_Barrier(self):
        comm.Barrier()
        comm.barrier()

    def test_Bcast(self):
        comm.Bcast(self.array, self.buffer)
        assert (self.buffer == self.array).all()
        assert (comm.bcast(self.array) == self.array).all()

    def test_Gather(self):
        comm.Gather(self.array, self.buffer)
        assert (self.buffer == self.array).all()
        assert (comm.gather(self.array)[0] == self.array).all()
        comm.Gatherv(self.array, self.buffer)
        assert (self.buffer == self.array).all()

    def test_Is_intra(self):
        assert comm.Is_intra()

    def test_Is_inter(self):
        assert not comm.Is_inter()

    def test_Reduce(self):
        comm.Reduce(self.array, self.buffer)
        assert (self.buffer == self.array).all()
        assert (comm.reduce(self.array) == self.array).all()
        assert (comm.reduce(1) == 1)

    def test_Scatter(self):
        comm.Scatter([self.array], self.buffer)
        assert (self.buffer == self.array).all()
        assert (comm.scatter([self.array]) == self.array).all()
        comm.Scatterv([self.array], self.buffer)
        assert (self.buffer == self.array).all()

    def test_Sendrecv(self):
        assert (comm.Sendrecv(self.array) == self.array).all()
        assert (comm.sendrecv(self.array) == self.array).all()



# -*- coding: utf-8 -*-

# %% IMPORTS
# mpi4pyd imports
from mpi4pyd.utils import rprint


# %% PYTEST CLASSES AND FUNCTIONS
# Pytest for the rprint function
def test_rprint():
    # Check if rprint works correctly
    rprint('Testing')



# -*- coding: utf-8 -*-

"""
mpi4pyd
=======
MPI for Python Dummies.

"""


# %% IMPORTS
# mpi4pyd imports
from .__version__ import __version__
from . import dummyMPI
from . import MPI
from .MPI import get_BufferComm_obj
from . import utils
from .utils import *

# All declaration
__all__ = ['dummyMPI', 'MPI', 'utils', 'get_BufferComm_obj']
__all__.extend(utils.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
Dummy MPI
=========
Dummy module that emulates the functionality of the :mod:`mpi4py.MPI` module.
It is automatically imported as :mod:`mpi4pyd.MPI` if :mod:`mpi4py.MPI` is not
available.
This is a specialized version of the `mpi_dummy` package available at
https://gitlab.mpcdf.mpg.de/ift/mpi_dummy

"""


# %% IMPORTS
# Built-in imports
from copy import deepcopy as copy
from pkg_resources import parse_version

# Package imports
import numpy as np

# mpi4pyd imports
from mpi4pyd import __version__

# All declaration
__all__ = ['COMM_SELF', 'COMM_WORLD', 'Comm', 'Datatype', 'Intracomm', 'Op',
           'AINT', 'BAND', 'BOOL', 'BOR', 'BXOR', 'BYTE', 'CHAR', 'CHARACTER',
           'COMPLEX', 'COMPLEX16', 'COMPLEX32', 'COMPLEX4', 'COMPLEX8',
           'COUNT', 'CXX_BOOL', 'CXX_DOUBLE_COMPLEX', 'CXX_FLOAT_COMPLEX',
           'CXX_LONG_DOUBLE_COMPLEX', 'C_BOOL', 'C_COMPLEX',
           'C_DOUBLE_COMPLEX', 'C_FLOAT_COMPLEX', 'C_LONG_DOUBLE_COMPLEX',
           'DATATYPE_NULL', 'DOUBLE', 'DOUBLE_COMPLEX', 'DOUBLE_INT',
           'DOUBLE_PRECISION', 'FLOAT', 'FLOAT_INT', 'F_BOOL', 'F_COMPLEX',
           'F_DOUBLE', 'F_DOUBLE_COMPLEX', 'F_FLOAT', 'F_FLOAT_COMPLEX',
           'F_INT', 'INT', 'INT16_T', 'INT32_T', 'INT64_T', 'INT8_T',
           'INTEGER', 'INTEGER1', 'INTEGER16', 'INTEGER2', 'INTEGER4',
           'INTEGER8', 'INT_INT', 'LAND', 'LB', 'LOGICAL', 'LOGICAL1',
           'LOGICAL2', 'LOGICAL4', 'LOGICAL8', 'LONG', 'LONG_DOUBLE',
           'LONG_DOUBLE_INT', 'LONG_INT', 'LONG_LONG', 'LOR', 'LXOR', 'OFFSET',
           'MAX', 'MAXLOC', 'MIN', 'MINLOC', 'NO_OP', 'OP_NULL', 'PACKED',
           'PROD', 'REAL', 'REAL16', 'REAL2', 'REAL4', 'REAL8', 'REPLACE',
           'SHORT', 'SHORT_INT', 'SIGNED_CHAR', 'SIGNED_INT', 'SIGNED_LONG',
           'SIGNED_LONG_LONG', 'SIGNED_SHORT', 'SINT16_T', 'SINT32_T',
           'SINT64_T', 'SINT8_T', 'SUM', 'TWOINT', 'UB', 'UINT16_T',
           'UINT32_T', 'UINT64_T', 'UINT8_T', 'UNSIGNED', 'UNSIGNED_CHAR',
           'UNSIGNED_INT', 'UNSIGNED_LONG', 'UNSIGNED_LONG_LONG',
           'UNSIGNED_SHORT', 'WCHAR', 'get_vendor']


# %% COMM CLASS DEFINITION
# Make dummy Comm class
class Comm(object):
    def __init__(self):
        # Save name of this class if not saved already
        if not hasattr(self, '_name'):
            self.name = "dummyMPI_%s" % (self.__class__.__name__)
        # If it was saved, check that it starts with 'dummyMPI_'
        elif not self._name.startswith('dummyMPI_'):
            # If not, add it
            self._name = "dummyMPI_%s" % (self._name)

        # Save rank and size of communicator
        self._rank = 0
        self._size = 1

    # %% CLASS PROPERTIES
    @property
    def name(self):
        return(self._name)

    @name.setter
    def name(self, name):
        if isinstance(name, str):
            self._name = name
        else:
            raise TypeError("Input argument 'name' is not of type 'str'!")

    @property
    def rank(self):
        return(self._rank)

    @property
    def size(self):
        return(self._size)

    # %% GENERAL CLASS METHODS
    def _get_buffer(self, buff):
        # If buff is a list or tuple, return the first element
        if isinstance(buff, (list, tuple)):
            return(buff[0])
        # Else, return buff itself
        else:
            return(buff)

    def _scatter_gather(self, sendbuf, recvbuf=None):
        # Unwrap the sending and receiving buffers
        sendbuf = self._get_buffer(sendbuf)
        recvbuf = self._get_buffer(recvbuf)

        # If no receiving buffer was supplied, return a copy of sendbuf
        if recvbuf is None:
            # Make a copy of sendbuf
            return(copy(sendbuf))
        # If a receiving buffer was supplied, use it
        else:
            recvbuf[:] = sendbuf
            return(recvbuf)

    # %% VISIBLE CLASS METHODS
    # TODO: Implement dummy versions of missing communication methods
    # Still missing: Alltoall and non-blocking/synchronous (I/S) methods
    def Get_name(self):
        return(self.name)

    def Get_rank(self):
        return(self.rank)

    def Get_size(self):
        return(self.size)

    def Allgather(self, sendbuf, recvbuf, *args, **kwargs):
        return(self.Gather(sendbuf, recvbuf))

    def allgather(self, sendobj, *args, **kwargs):
        return(self.gather(sendobj))

    def Allgatherv(self, sendbuf, recvbuf, *args, **kwargs):
        return(self.Gatherv(sendbuf, recvbuf))

    def Allreduce(self, sendbuf, recvbuf, *args, **kwargs):
        return(self.Reduce(sendbuf, recvbuf))

    def allreduce(self, sendobj, *args, **kwargs):
        return(self.reduce(sendobj))

    def Barrier(self):
        pass

    def barrier(self):
        pass

    def Bcast(self, buf, *args, **kwargs):
        return(buf)

    def bcast(self, obj, *args, **kwargs):
        return(obj)

    def Gather(self, sendbuf, recvbuf, *args, **kwargs):
        return(self._scatter_gather(sendbuf, recvbuf))

    def gather(self, sendobj, *args, **kwargs):
        return([sendobj])

    def Gatherv(self, sendbuf, recvbuf, *args, **kwargs):
        return(self._scatter_gather(sendbuf, recvbuf))

    def Is_intra(self):
        return(isinstance(self, Intracomm))

    def Is_inter(self):
        return(False)

    def Reduce(self, sendbuf, recvbuf, *args, **kwargs):
        return(self._scatter_gather(sendbuf, recvbuf))

    def reduce(self, sendobj, *args, **kwargs):
        if np.isscalar(sendobj):
            return(sendobj)
        else:
            return(self._scatter_gather(sendobj))

    def Scatter(self, sendbuf, recvbuf, *args, **kwargs):
        return(self._scatter_gather(sendbuf, recvbuf))

    def scatter(self, sendobj, *args, **kwargs):
        return(sendobj[0])

    def Scatterv(self, sendbuf, recvbuf, *args, **kwargs):
        return(self._scatter_gather(sendbuf, recvbuf))

    def Sendrecv(self, sendbuf, *args, **kwargs):
        return(sendbuf)

    def sendrecv(self, sendobj, *args, **kwargs):
        return(sendobj)


# %% INTRACOMM CLASS DEFINITION
# Make dummy Intracomm class
class Intracomm(Comm):
    def __init__(self, name, *args, **kwargs):
        self.name = name
        super(Intracomm, self).__init__(*args, **kwargs)


# %% INITIALIZE COMM_WORLD AND COMM_SELF
COMM_WORLD = Intracomm('dummyMPI_COMM_WORLD')
COMM_SELF = Intracomm('dummyMPI_COMM_SELF')


# %% DATATYPE DEFINITIONS
# Make dummy Datatype class
class Datatype(object):
    def __init__(self, name):
        self.name = name


# MPI standard datatypes
AINT = Datatype('dummyMPI_AINT')
BYTE = Datatype('dummyMPI_BYTE')
CHAR = Datatype('dummyMPI_CHAR')
CHARACTER = Datatype('dummyMPI_CHARACTER')
COMPLEX = Datatype('dummyMPI_COMPLEX')
COMPLEX4 = Datatype('dummyMPI_COMPLEX4')
COMPLEX8 = Datatype('dummyMPI_COMPLEX8')
COMPLEX16 = Datatype('dummyMPI_COMPLEX16')
COMPLEX32 = Datatype('dummyMPI_COMPLEX32')
COUNT = Datatype('dummyMPI_COUNT')
CXX_BOOL = Datatype('dummyMPI_CXX_BOOL')
CXX_DOUBLE_COMPLEX = Datatype('dummyMPI_CXX_DOUBLE_COMPLEX')
CXX_FLOAT_COMPLEX = Datatype('dummyMPI_CXX_FLOAT_COMPLEX')
CXX_LONG_DOUBLE_COMPLEX = Datatype('dummyMPI_CXX_LONG_DOUBLE_COMPLEX')
C_BOOL = Datatype('dummyMPI_C_BOOL')
C_COMPLEX = Datatype('dummyMPI_C_COMPLEX')
C_DOUBLE_COMPLEX = Datatype('dummyMPI_C_DOUBLE_COMPLEX')
C_FLOAT_COMPLEX = Datatype('dummyMPI_C_FLOAT_COMPLEX')
C_LONG_DOUBLE_COMPLEX = Datatype('dummyMPI_C_LONG_DOUBLE_COMPLEX')
DATATYPE_NULL = Datatype('dummyMPI_DATATYPE_NULL')
DOUBLE = Datatype('dummyMPI_DOUBLE')
DOUBLE_COMPLEX = Datatype('dummyMPI_DOUBLE_COMPLEX')
DOUBLE_INT = Datatype('dummyMPI_DOUBLE_INT')
DOUBLE_PRECISION = Datatype('dummyMPI_DOUBLE_PRECISION')
FLOAT = Datatype('dummyMPI_FLOAT')
FLOAT_INT = Datatype('dummyMPI_FLOAT_INT')
INT = Datatype('dummyMPI_INT')
INT8_T = Datatype('dummyMPI_INT8_T')
INT16_T = Datatype('dummyMPI_INT16_T')
INT32_T = Datatype('dummyMPI_INT32_T')
INT64_T = Datatype('dummyMPI_INT64_T')
INTEGER = Datatype('dummyMPI_INTEGER')
INTEGER1 = Datatype('dummyMPI_INTEGER1')
INTEGER2 = Datatype('dummyMPI_INTEGER2')
INTEGER4 = Datatype('dummyMPI_INTEGER4')
INTEGER8 = Datatype('dummyMPI_INTEGER8')
INTEGER16 = Datatype('dummyMPI_INTEGER16')
INT_INT = Datatype('dummyMPI_2INT')
LB = Datatype('dummyMPI_LB')
LOGICAL = Datatype('dummyMPI_LOGICAL')
LOGICAL1 = Datatype('dummyMPI_LOGICAL1')
LOGICAL2 = Datatype('dummyMPI_LOGICAL2')
LOGICAL4 = Datatype('dummyMPI_LOGICAL4')
LOGICAL8 = Datatype('dummyMPI_LOGICAL8')
LONG = Datatype('dummyMPI_LONG')
LONG_DOUBLE = Datatype('dummyMPI_LONG_DOUBLE')
LONG_DOUBLE_INT = Datatype('dummyMPI_LONG_DOUBLE_INT')
LONG_INT = Datatype('dummyMPI_LONG_INT')
LONG_LONG = Datatype('dummyMPI_LONG_LONG_INT')
OFFSET = Datatype('dummyMPI_OFFSET')
PACKED = Datatype('dummyMPI_PACKED')
REAL = Datatype('dummyMPI_REAL')
REAL2 = Datatype('dummyMPI_REAL2')
REAL4 = Datatype('dummyMPI_REAL4')
REAL8 = Datatype('dummyMPI_REAL8')
REAL16 = Datatype('dummyMPI_REAL16')
SHORT = Datatype('dummyMPI_SHORT')
SHORT_INT = Datatype('dummyMPI_SHORT_INT')
SIGNED_CHAR = Datatype('dummyMPI_SIGNED_CHAR')
UB = Datatype('dummyMPI_UB')
UINT8_T = Datatype('dummyMPI_UINT8_T')
UINT16_T = Datatype('dummyMPI_UINT16_T')
UINT32_T = Datatype('dummyMPI_UINT32_T')
UINT64_T = Datatype('dummyMPI_UINT64_T')
UNSIGNED = Datatype('dummyMPI_UNSIGNED')
UNSIGNED_CHAR = Datatype('dummyMPI_UNSIGNED_CHAR')
UNSIGNED_LONG = Datatype('dummyMPI_UNSIGNED_LONG')
UNSIGNED_LONG_LONG = Datatype('dummyMPI_UNSIGNED_LONG_LONG')
UNSIGNED_SHORT = Datatype('dummyMPI_UNSIGNED_SHORT')
WCHAR = Datatype('dummyMPI_WCHAR')

# MPI datatype synonyms
BOOL = C_BOOL
F_BOOL = LOGICAL
F_COMPLEX = COMPLEX
F_DOUBLE = DOUBLE_PRECISION
F_DOUBLE_COMPLEX = DOUBLE_COMPLEX
F_FLOAT = REAL
F_FLOAT_COMPLEX = COMPLEX
F_INT = INTEGER
SIGNED_INT = INT
SIGNED_LONG = LONG
SIGNED_LONG_LONG = LONG_LONG
SIGNED_SHORT = SHORT
SINT8_T = INT8_T
SINT16_T = INT16_T
SINT32_T = INT32_T
SINT64_T = INT64_T
TWOINT = INT_INT
UNSIGNED_INT = UNSIGNED


# %% OPERATOR DEFINITIONS
# Make dummy Op class
class Op(object):
    def __init__(self, *args, **kwargs):
        self.is_predefined = True

    def __call__(self, *args, **kwargs):
        pass


# MPI standard operators
BAND = Op()
BOR = Op()
BXOR = Op()
LAND = Op()
LOR = Op()
LXOR = Op()
MAX = Op()
MAXLOC = Op()
MIN = Op()
MINLOC = Op()
NO_OP = Op()
OP_NULL = Op()
PROD = Op()
REPLACE = Op()
SUM = Op()


# %% DUMMY FUNCTIONS
def get_vendor():
    return("dummyMPI", parse_version(__version__)._version.release)



# -*- coding: utf-8 -*-

"""
BufferComm
==========

"""


# %% IMPORTS
# Built-in imports
from inspect import currentframe

# Package imports
from e13tools import InputError, ShapeError
import numpy as np

# mpi4pyd imports
from mpi4pyd import dummyMPI, MPI
from mpi4pyd.MPI._helpers import is_buffer_obj

# All declaration
__all__ = ['BUFFER_COMM_SELF', 'BUFFER_COMM_WORLD', 'get_BufferComm_obj']


# Initialize buffer_comm_registry
buffer_comm_registry = {}

# Make conversion dict from NumPy dtype to MPI Datatype
dtype_dict = {
    'int': MPI.LONG,
    'int32': MPI.INT,
    'int64': MPI.LONG,
    'float': MPI.DOUBLE,
    'float32': MPI.FLOAT,
    'float64': MPI.DOUBLE}


# %% FUNCTION DEFINITIONS
# Function factory that returns special BufferComm class instances
def get_BufferComm_obj(comm=None):
    """
    Function factory that returns an instance of the :class:`~BufferComm`
    class, defined as ``BufferComm(comm.__class__, object)``.

    This :class:`~BufferComm` class wraps the provided :obj:`MPI.Intracomm`
    instance `comm` and overrides all of its lowercase communication methods
    (e.g., :meth:`~MPI.Intracomm.bcast`, :meth:`~MPI.Intracomm.gather`,
    :meth:`~MPI.Intracomm.scatter`, :meth:`~MPI.Intracomm.recv` and
    :meth:`~MPI.Intracomm.send`) with improved versions. These improved
    communication methods automatically select the most optimal way of
    communicating their input arguments.

    Besides the new method functionalities, the returned instance behaves in
    the exact same way as the provided `comm` and can easily be used in any
    algorithm that expects an instance of the :class:`MPI.Intracomm` class.

    Optional
    --------
    comm : :obj:`~MPI.Intracomm` object or None. Default: None
        The MPI intra-communicator to use as the base for the
        :obj:`~BufferComm` instance.
        If *None*, use :obj:`MPI.COMM_WORLD` instead.

    Returns
    -------
    buffer_comm : :obj:`MPI.Comm` object
        The provided `comm` which has its lowercase communication methods
        overridden. If `comm` is *None* or :obj:`MPI.COMM_WORLD`,
        :obj:`mpi4pyd.MPI.BUFFER_COMM_WORLD` is returned instead.

    Note
    ----
    Providing the same :obj:`~MPI.Intracomm` instance to this function twice,
    will not create two :obj:`~BufferComm` objects. Instead, the instance
    created the first time will be returned each consecutive time. All created
    :obj:`~BufferComm` objects are stored in the :obj:`~buffer_comm_registry`.

    """

    # If comm is None, set it to MPI.COMM_WORLD
    if comm is None:
        comm = MPI.COMM_WORLD
    # Else, check if provided comm is an MPI intra-communicator
    elif not isinstance(comm, (MPI.Intracomm, dummyMPI.Intracomm)):
        raise TypeError("Input argument 'comm' must be an instance of "
                        "the MPI.Intracomm class!")

    # Check if provided comm already has a BufferComm instance
    if hex(id(comm)) in buffer_comm_registry.keys():
        # If so, return that BufferComm instance instead
        return(buffer_comm_registry[hex(id(comm))])

    # Check if provided comm is not already a BufferComm instance
    if comm in buffer_comm_registry.values():
        # If so, return provided BufferComm instance instead
        return(comm)

    # Make tuple of overridden attributes
    overridden_attrs = ('__init__', 'bcast', 'gather', 'scatter')

    # %% BUFFERCOMM CLASS DEFINITION
    class BufferComm(comm.__class__, object):
        """
        Custom :class:`~MPI.Intracomm` class.

        """

        def __init__(self):
            # Bind provided communicator
            if not hasattr(self, '_rank'):
                self._rank = comm.Get_rank()
            if not hasattr(self, '_size'):
                self._size = comm.Get_size()

        # If requested attribute is not a method, use comm for getattr
        def __getattribute__(self, name):
            if name in dir(comm) and name not in overridden_attrs:
                return(getattr(comm, name))
            else:
                return(super().__getattribute__(name))

        # If requested attribute is not a method, use comm for setattr
        def __setattr__(self, name, value):
            if name in dir(comm) and name not in overridden_attrs:
                setattr(comm, name, value)
            else:
                super().__setattr__(name, value)

        # If requested attribute is not a method, use comm for delattr
        def __delattr__(self, name):
            if name in dir(comm) and name not in overridden_attrs:
                delattr(comm, name)
            else:
                super().__delattr__(name)

        # %% COMMUNICATION METHODS
        # Specialized bcast function that automatically makes use of buffers
        def bcast(self, obj, root=0):
            """
            Special broadcast method that automatically uses the appropriate
            method (:meth:`~MPI.Intracomm.bcast` or
            :meth:`~MPI.Intracomm.Bcast`) depending on the type of the provided
            `obj`.

            Parameters
            ----------
            obj : :obj:`~numpy.ndarray` or object
                The object to broadcast to all MPI ranks.
                If :obj:`~numpy.ndarray`, use :meth:`~MPI.Intracomm.Bcast`.
                If not, use :meth:`~MPI.Intracomm.bcast` instead.

            Optional
            --------
            root : int. Default: 0
                The MPI rank that broadcasts `obj`.

            Returns
            -------
            obj : object
                The broadcasted `obj`.

            """

            # Check if obj can be broadcasted as a buffer object
            use_buffer = self.__use_buffer_meth(obj, root)

            # If provided object uses a buffer
            if use_buffer:
                # Sender
                if(self._rank == root):
                    # If so, send shape and dtype of the NumPy array
                    comm.bcast([obj.shape, obj.dtype], root=root)

                    # Then send the NumPy array as a buffer object
                    comm.Bcast(obj, root=root)

                # Receivers receive NumPy array
                else:
                    # Create empty NumPy array with given shape and dtype
                    obj = np.empty(*comm.bcast(None, root=root))

                    # Receive NumPy array
                    comm.Bcast(obj, root=root)

            # If not, broadcast obj the normal way
            else:
                # Try to broadcast object
                try:
                    obj = comm.bcast(obj, root=root)
                # If this fails, raise error about byte size
                except OverflowError:
                    raise InputError("Input argument `obj` has a byte size "
                                     "that cannot be stored in a 32-bit int "
                                     "(%i > %i)!"
                                     % (obj.__sizeof__(), 2**31-1))

            # Return obj
            return(obj)

        # Specialized gather function that automatically makes use of buffers
        def gather(self, obj, root=0):
            """
            Special gather method that automatically uses the appropriate
            method (:meth:`~MPI.Intracomm.gather` or
            :meth:`~MPI.Intracomm.Gatherv`) depending on the lay-out of the
            provided `obj`.

            Parameters
            ----------
            obj : :obj:`~numpy.ndarray` or object
                The object to gather from all MPI ranks.
                If :obj:`~numpy.ndarray`, use :meth:`~MPI.Intracomm.Gatherv`.
                If not, use :meth:`~MPI.Intracomm.gather` instead.

            Optional
            --------
            root : int. Default: 0
                The MPI rank that gathers `obj`.

            Returns
            -------
            obj : list, object or None
                If MPI rank is `root`, returns a list of gathered objects or a
                single object, depending on `default`.
                Else, returns *None*.

            Warnings
            --------
            When gathering NumPy arrays, all arrays must have the same number
            of dimensions and the same shape, except for one axis.

            """

            # Check if obj can be gathered as a buffer object
            use_buffer = self.__use_buffer_meth(obj, root)

            # If all provided objects use buffers
            if use_buffer:
                # If so, gather the shapes of obj on the receiver
                shapes = np.array(comm.gather(obj.shape, root=root))

                # Receiver sets up a buffer array and receives NumPy array
                if(self._rank == root):
                    # Obtain counts and displacements
                    counts = np.product(shapes, axis=1)
                    disps = np.cumsum([0, *counts[:-1]])

                    # Get the maximum size in every axis
                    max_size = np.max(shapes, axis=0)

                    # Check which axis sizes differ
                    diff_size = ~np.all(np.equal(shapes, max_size), axis=0)

                    # If more than a single axis size differs, raise error
                    # TODO: Remove this limitation
                    if(sum(diff_size) > 1):
                        raise ShapeError("Input argument 'obj' differs in size"
                                         "in more than 1 axis!")

                    # Get the axis that differs and its cumulative size
                    diff_axis =\
                        np.nonzero(diff_size)[0][0] if sum(diff_size) else 0

                    # Get the buffer shape and size
                    buff_shape = max_size
                    buff_shape[diff_axis] = np.sum(shapes[:, diff_axis])
                    buff_size = np.product(buff_shape)

                    # Initialize empty buffer array
                    recv_obj = np.empty(buff_size, dtype=obj.dtype)

                    # Make buffer list
                    buff =\
                        [recv_obj, counts, disps, dtype_dict[obj.dtype.name]]

                    # Gather all NumPy arrays
                    comm.Gatherv([obj.ravel(), obj.size], buff, root=root)

                    # Reconstruct the original shapes of NumPy arrays
                    arr_list = np.split(recv_obj, disps[1:])
                    for i, shape in enumerate(shapes):
                        arr_list[i] = np.reshape(arr_list[i], shape)
                    recv_obj = arr_list

                # Senders send the array
                else:
                    # Send NumPy array
                    comm.Gatherv([obj, obj.size], None, root=root)
                    recv_obj = None

            # If not, gather obj the normal way
            else:
                # Try to gather the obj
                try:
                    recv_obj = comm.gather(obj, root=root)
                # If this fails, raise error about byte size
                except SystemError:
                    raise InputError("Input argument 'obj' is too large!")

            # Return recv_obj
            return(recv_obj)

        # Specialized scatter function that automatically makes use of buffers
        def scatter(self, obj, root=0):
            """
            Special scatter method that automatically uses the appropriate
            method (:meth:`~MPI.Intracomm.scatter` or
            :meth:`~MPI.Intracomm.Scatter`) depending on the type of the
            provided `obj`.

            Unlike :meth:`~MPI.Intracomm.scatter`, providing a buffer object
            with more than :attr:`~_size` items will not raise an error, but
            evenly distribute all the items instead.

            Parameters
            ----------
            obj : :obj:`~numpy.ndarray` or object
                The object to scatter to all MPI ranks.
                If :obj:`~numpy.ndarray`, use :meth:`~MPI.Intracomm.Scatterv`.
                If not, use :meth:`~MPI.Intracomm.scatter` instead.

            Optional
            --------
            root : int. Default: 0
                The MPI rank that scatters `obj`.

            Returns
            -------
            obj : :obj:`~numpy.ndarray` or object
                The object that has been scattered to this MPI rank.

            """

            # Check if obj can be scattered as buffer objects
            use_buffer = self.__use_buffer_meth(obj, root)

            # If provided object uses a buffer
            if use_buffer:
                # Sender prepares for scattering
                if(self._rank == root):
                    # Raise error if length of axis is not divisible by size
                    if len(obj) % self._size:
                        raise ShapeError("Input argument 'obj' cannot be "
                                         "divided evenly over the available "
                                         "number of MPI ranks!")

                    # Determine shape of scattered object
                    buff_shape = list(obj.shape)
                    buff_shape[0] //= self._size

                    # Initialize empty buffer array
                    recv_obj = np.empty(*comm.bcast([buff_shape, obj.dtype],
                                                    root=root))

                    # Scatter NumPy array
                    comm.Scatter(obj, recv_obj, root=root)

                # Receivers receive the array
                else:
                    # Initialize empty buffer array
                    recv_obj = np.empty(*comm.bcast(None, root=root))

                    # Receive scattered NumPy array
                    comm.Scatter(None, recv_obj, root=root)

                # Remove single dimensional entries from recv_obj
                recv_obj = recv_obj.squeeze()

            # If not, scatter obj the normal way
            else:
                # Try to scatter the obj
                try:
                    recv_obj = comm.scatter(obj, root=root)
                # If this fails, raise error about byte size
                except SystemError:
                    raise InputError("Input argument 'obj' is too large!")

            # Return recv_obj
            return(recv_obj)

        # %% UTILITY METHODS
        # This function checks if a buffer communication method can be used
        def __use_buffer_meth(self, obj, root):
            """
            Depending on which communication method calls this function,
            determines if the provided `obj` on all MPI ranks can be
            communicated using an uppercase communication method.

            This method must be called by all MPI ranks.
            This method must never be called directly.

            """

            # Determine the name of the frame calling this method
            meth_name = currentframe().f_back.f_code.co_name

            # Check who called this method and act accordingly
            if meth_name in ('bcast', 'scatter'):
                return(comm.bcast(is_buffer_obj(obj), root=root))
            elif meth_name in ('gather'):
                return(comm.allreduce(is_buffer_obj(obj), op=MPI.MIN))
            else:
                raise NotImplementedError

    # %% REMAINDER OF FUNCTION FACTORY
    # Initialize BufferComm
    buffer_comm = BufferComm()

    # Register initialized BufferComm
    buffer_comm_registry[hex(id(comm))] = buffer_comm

    # Return buffer_comm
    return(buffer_comm)


# %% DEFAULT INSTANCES
BUFFER_COMM_SELF = get_BufferComm_obj(MPI.COMM_SELF)
BUFFER_COMM_WORLD = get_BufferComm_obj(MPI.COMM_WORLD)



# -*- coding: utf-8 -*-

"""
mpi4pyd Setup
=============
Contains the setup script required for installing the *mpi4pyd* package.
This can be ran directly by using::

    pip install .

or anything equivalent.

"""


# %% IMPORTS
# Built-in imports
from codecs import open
import re

# Package imports
from setuptools import find_packages, setup


# %% SETUP DEFINITION
# Get the long description from the README file
with open('README.rst', 'r') as f:
    long_description = f.read()

# Get the requirements list by reading the file and splitting it up
with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()

# Read the __version__.py file
with open('mpi4pyd/__version__.py', 'r') as f:
    vf = f.read()

# Obtain version from read-in __version__.py file
version = re.search(r"^_*version_* = ['\"]([^'\"]*)['\"]", vf, re.M).group(1)

# Setup function declaration
setup(name='mpi4pyd',
      version=version,
      author="Ellert van der Velden",
      author_email="ellert_vandervelden@outlook.com",
      maintainer="1313e",
      description=("mpi4pyd: MPI for Python Dummies"),
      long_description=long_description,
      url="https://github.com/1313e/mpi4pyd",
      license='BSD-3',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Operating System :: MacOS',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: Unix',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Software Development'
          ],
      keywords=('mpi4pyd'),
      python_requires='>=3.5, <4',
      packages=find_packages(),
      package_dir={'mpi4pyd': "mpi4pyd"},
      include_package_data=True,
      install_requires=requirements,
      zip_safe=False,
      )



# -*- coding: utf-8 -*-

"""
MPI
===
Module that automatically picks the correct MPI module based on whether or not
the :mod:`mpi4py.MPI` module is available.

"""


# %% IMPORTS
# MPI import
try:
    from mpi4py import MPI as _MPI
    from mpi4py.MPI import *
except ImportError:
    from mpi4pyd import dummyMPI as _MPI
    from mpi4pyd.dummyMPI import *
from . import _buffer_comm
from ._buffer_comm import *

# All declaration
__all__ = []
if(_MPI.__package__ == 'mpi4py'):
    __all__.extend([prop for prop in dir(_MPI) if not prop.startswith('_')])
else:
    __all__.extend(_MPI.__all__)
__all__.extend(_buffer_comm.__all__)

# Name and package declaration
__name__ = getattr(_MPI, '__name__', None)
__package__ = getattr(_MPI, '__package__', None)
